/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle;

import static de.ozgcloud.formcycle.errorhandling.NodeThrewExceptionFactory.*;
import static org.apache.commons.lang3.StringUtils.*;

import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.Supplier;

import com.fasterxml.jackson.databind.ObjectMapper;

import de.ozgcloud.formcycle.attachment.AttachmentMapper;
import de.ozgcloud.formcycle.client.OzgCloudResponseHandler;
import de.ozgcloud.formcycle.errorhandling.OzgPluginError;
import de.ozgcloud.formcycle.errorhandling.OzgPluginSoftError;
import de.ozgcloud.formcycle.errorhandling.NodeThrewExceptionFactory;
import de.ozgcloud.formcycle.errorhandling.OzgCloudRequestException;
import de.ozgcloud.formcycle.errorhandling.TechnicalException;
import de.ozgcloud.formcycle.formdata.OzgCloudFormDataMapper;
import de.ozgcloud.formcycle.formdata.PluginFormDataAdapter;
import de.xima.fc.beans.interfaces.IGuiIcon;
import de.xima.fc.entities.Attachment;
import de.xima.fc.exceptions.AbstractAbruptCompletionException;
import de.xima.fc.interfaces.plugin.lifecycle.IPluginInitializeData;
import de.xima.fc.interfaces.workflow.execution.IFileValueDescriptor;
import de.xima.fc.interfaces.workflow.execution.INormalCompletionResult;
import de.xima.fc.interfaces.workflow.mixin.IElementCategory;
import de.xima.fc.interfaces.workflow.mixin.IKeyValueSummarizableNode;
import de.xima.fc.interfaces.workflow.mixin.ISummaryKeyValueModel;
import de.xima.fc.interfaces.workflow.params.IGetElementPrototypesParams;
import de.xima.fc.interfaces.workflow.params.IGetElementSummaryParams;
import de.xima.fc.interfaces.workflow.params.INodeExecutionParams;
import de.xima.fc.interfaces.workflow.value.IUnionValueDescriptor;
import de.xima.fc.interfaces.workflow.value.IValueBuilder;
import de.xima.fc.interfaces.workflow.value.IValueDescriptor;
import de.xima.fc.interfaces.workflow.value.IValueDescriptorFactory;
import de.xima.fc.mdl.ui.GuiIcon;
import de.xima.fc.workflow.FileValueDescriptor;
import de.xima.fc.workflow.SummaryKeyValueModelBuilder;
import de.xima.fc.workflow.enums.EWorkflowElementCategory;
import de.xima.fc.workflow.mixin.IPluginActionNodeHandler;

public final class OzgCloudPlugin
		implements IPluginActionNodeHandler<OzgPluginWorkflowNodeProperties>, IKeyValueSummarizableNode<OzgPluginWorkflowNodeProperties> {

	static final String VORGANGNUMMER_PROPERTY_KEY = "vorgangnummer";
	public static final String KEY_PLUGIN_VERSION = "pluginVersion";

	private final transient PortParser portParser = new PortParser();
	private final transient BuildProperties buildProperties = new BuildProperties();
	private transient OzgPluginPropertiesSupplier ozgPluginPropertiesSupplier;
	private transient IPluginInitializeData initializeData;
	private String pluginVersion;

	@Override
	public void initialize(IPluginInitializeData initializeData) {
		this.initializeData = initializeData;
		this.ozgPluginPropertiesSupplier = new OzgPluginPropertiesSupplier(initializeData, portParser);
		this.pluginVersion = buildProperties.readPluginVersion();
	}

	@Override
	public String getPropertiesViewXhtmlName() {
		return "OzgCloudPlugin.xhtml";
	}

	@Override
	public String getName() {
		return "de.ozgcloud.formcycle.OzgCloudPlugin";
	}

	@Override
	public IElementCategory getMainCategory(IGetElementPrototypesParams params) {
		// Main category where this node is shown in the drawer panel to the left of the
		// designer.
		return EWorkflowElementCategory.MAIN_ACTION;
	}

	@Override
	public IElementCategory getSubCategory(IGetElementPrototypesParams params) {
		// Sub category where this node is shown in the drawer panel to the left of the
		// designer.
		return EWorkflowElementCategory.SUB_ACTION_COMMON;
	}

	@Override
	public IGuiIcon getPrototypeIcon(IGetElementPrototypesParams params) {
		// Optional icon for this workflow node. Shown next to the node's name.
		return GuiIcon.valueOf("fa-question");
	}

	@Override
	public IFileValueDescriptor getFileValueDescriptor() {
		// Implement this method when this node can return files.
		// This node can then be selected by other actions that require files.
		// If this node does not provide files, remove this method
		return new FileValueDescriptor("OzgCloudPlugin.files");
	}

	@Override
	public IValueDescriptor<?, ? extends IValueBuilder<?>> getSuccessValueDescriptor(IValueDescriptorFactory factory) {
		// Data returned by this node when it is executed
		// Users can access this data e.g. via placeholders, for example:
		// [%$Action.RESULT.name%]
		// [%$Action.RESULT.items.name%]
		// [%$Action.RESULT.items.state%]
		return factory.recordBuilder().requiredProperty(VORGANGNUMMER_PROPERTY_KEY, factory.string()).build();
	}

	@Override
	public IUnionValueDescriptor<String> getErrorValueDescriptor(IValueDescriptorFactory factory) {
		// For each error that can occur, additional data can be provided
		// Users can access this data e.g. via placeholders, for example:
		// [%$Action.ERROR_CODE%]
		// [%$Action.ERROR_MESSAGE%]
		// [%$Action.ERROR.file%]
		// [%$Action.ERROR.requiredPermissions[0]%]
		// [%$Action.ERROR.message%]
		return factory.unionStringBuilder() //
				.add(OzgPluginError.INTERNAL_ERROR.name(),
						factory.recordBuilder().requiredProperty(KEY_EXCEPTION_ID, factory.string()).build())
				.add(OzgPluginError.OZGCLOUD_REQUEST_FAILED.name(),
						factory.recordBuilder().requiredProperty(KEY_EXCEPTION_ID, factory.string()).build())
				.build();
	}

	@Override
	public ISummaryKeyValueModel getElementSummaryKeyValueModel(IGetElementSummaryParams<OzgPluginWorkflowNodeProperties> params) {
		// A simple summary table with the most important fields from the properties
		// This is shown when the user clicks on the info icon on a node in the designer.
		return new SummaryKeyValueModelBuilder().build();
	}

	@Override
	public INormalCompletionResult execute(INodeExecutionParams<OzgPluginWorkflowNodeProperties> params) throws AbstractAbruptCompletionException {
		try {
			return buildSuccessResult(createPluginExecutor(params).execute(), params);
		} catch (OzgCloudRequestException e) {
			throw createExceptionFactory(params).createOzgCloudRequestException(e);
		} catch (TechnicalException e) {
			throw createExceptionFactory(params).createInternalException(e);
		} catch (Exception e) {
			throw createExceptionFactory(params).createInternalException("Unexpected error occurred", e);
		}
	}

	OzgPluginExecutor createPluginExecutor(INodeExecutionParams<OzgPluginWorkflowNodeProperties> params) {
		return new OzgPluginExecutor(new OzgCloudFormDataMapper(), createOzgCloudClient(params), new AttachmentMapper(),
				createFormDataAdapter(params), createAttachmentSupplier(params));
	}

	OzgCloudFormDataHttpClient createOzgCloudClient(INodeExecutionParams<OzgPluginWorkflowNodeProperties> params) {
		var ozgCloudConfig = buildOzgCloudConfig(params.getData());
		validateOzgCloudConfig(ozgCloudConfig);
		return new OzgCloudFormDataHttpClient(ozgCloudConfig, new SystemPropertiesProvider(), portParser,
				new OzgCloudResponseHandler(new ObjectMapper()));
	}

	void validateOzgCloudConfig(OzgCloudConfig ozgCloudConfig) {
		if (isBlank(ozgCloudConfig.getEingangsAdapterUrl())) {
			throw new TechnicalException("Eingangsadapter URL cannot be blank");
		}
	}

	PluginFormDataAdapter createFormDataAdapter(INodeExecutionParams<OzgPluginWorkflowNodeProperties> params) {
		return new PluginFormDataAdapter(params.getWorkflowContext().env(), params.getData().getOrganisationsEinheitId());
	}

	private Supplier<List<Attachment>> createAttachmentSupplier(INodeExecutionParams<OzgPluginWorkflowNodeProperties> params) {
		var formRecord = params.getWorkflowContext().env().getFormRecord();
		return () -> Optional.ofNullable(formRecord.getAttachments()).orElse(List.of());
	}

	NodeThrewExceptionFactory createExceptionFactory(INodeExecutionParams<OzgPluginWorkflowNodeProperties> params) {
		return new NodeThrewExceptionFactory(params.throwingException(), pluginVersion, buildOzgCloudConfig(params.getData()));
	}

	OzgCloudConfig buildOzgCloudConfig(OzgPluginWorkflowNodeProperties formProperties) {
		var ozgCloudConfigBuilder = OzgCloudConfig.builder().eingangsAdapterUrl(getEingangsAdapterUrl(formProperties));
		ozgPluginPropertiesSupplier.getProxyProperties().ifPresent(ozgCloudConfigBuilder::proxyConfig);
		return ozgCloudConfigBuilder.build();
	}

	String getEingangsAdapterUrl(OzgPluginWorkflowNodeProperties formProperties) {
		return formProperties.isUseCustomEingangsAdapterUrl()
				? formProperties.getEingangsAdapterUrl()
				: ozgPluginPropertiesSupplier.getEingangsAdapterUrl();
	}

	INormalCompletionResult buildSuccessResult(ExecutionResult pluginExecutionResult, INodeExecutionParams<OzgPluginWorkflowNodeProperties> params) {
		var resultBuilder = params.normalResult();
		resultBuilder.success(Map.of(VORGANGNUMMER_PROPERTY_KEY, pluginExecutionResult.getVorgangnummer()));
		pluginExecutionResult.getWarnings()
				.forEach(warning -> resultBuilder.softError(warning.getErrorCode(), warning.getMessage(), warning.getValue(), warning.getCause()));
		return resultBuilder.build();
	}

	@Override
	public Class<OzgPluginError> getErrorCodeClass() {
		// Class of the enumeration with the possible types of errors that may occur.
		return OzgPluginError.class;
	}

	@Override
	public Class<OzgPluginSoftError> getSoftErrorCodeClass() {
		// Class of the enumeration with the possible types of soft errors that may occur.
		return OzgPluginSoftError.class;
	}

	@Override
	public IPluginInitializeData getPluginInitializeData() {
		return initializeData;
	}

}
