/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle.client;

import java.io.IOException;
import java.nio.charset.StandardCharsets;

import org.apache.commons.io.IOUtils;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.ResponseHandler;

import com.fasterxml.jackson.databind.ObjectMapper;

import de.ozgcloud.eingang.formcycle.FormCycleConfirmationResponse;
import de.ozgcloud.formcycle.errorhandling.OzgCloudRequestException;
import lombok.RequiredArgsConstructor;

@RequiredArgsConstructor
public class OzgCloudResponseHandler implements ResponseHandler<FormCycleConfirmationResponse> {

	static final String MESSAGE = "Unexpected Status received: %s";

	private final ObjectMapper objectMapper;

	@Override
	public FormCycleConfirmationResponse handleResponse(HttpResponse response) throws IOException {
		checkResponseStatus(response);
		return buildResponse(response);
	}

	FormCycleConfirmationResponse buildResponse(HttpResponse response) throws IOException {
		return FormCycleConfirmationResponse.parseFrom(response.getEntity().getContent());
	}

	void checkResponseStatus(HttpResponse response) {
		var statusCode = response.getStatusLine().getStatusCode();
		if (HttpUtils.isError(statusCode)) {
			var receivedError = getOzgCloudError(response.getEntity());
			throw new OzgCloudRequestException(receivedError.getMessage(), receivedError.getExceptionId());
		}
		if (statusCode != HttpStatus.SC_OK) {
			throw new OzgCloudRequestException(buildFullMessage(response));
		}
	}

	OzgCloudErrorDto getOzgCloudError(HttpEntity entity) {
		try {
			return objectMapper.readValue(entity.getContent(), OzgCloudErrorDto.class);
		} catch (IOException e) {
			throw new OzgCloudRequestException("Cannot read error message.", e);
		}
	}

	String buildFullMessage(HttpResponse response) {
		StringBuilder messageBuilder = new StringBuilder(String.format(MESSAGE, response.getStatusLine()));
		try {
			IOUtils.readLines(response.getEntity().getContent(), StandardCharsets.UTF_8).forEach(line -> messageBuilder.append("\n").append(line));
		} catch (IOException e) {
			messageBuilder.append("\nError reading body");
		}
		return messageBuilder.toString();
	}
}
