/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle.formdata;

import static java.util.Objects.*;
import static org.apache.commons.lang3.StringUtils.*;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.function.Predicate;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;

import de.ozgcloud.formcycle.errorhandling.UnexpectedTrustLevelException;
import de.xima.fc.interfaces.form.IFormDataAdapter;
import de.xima.fc.interfaces.workflow.execution.IWorkflowExecutionEnvironmentData;
import de.xima.fc.mdl.enums.EFormFieldTyp;
import de.xima.fc.mdl.fdv.form.FormFieldMetaData;
import lombok.RequiredArgsConstructor;

@RequiredArgsConstructor
public class PluginFormDataAdapter {

	private static final Pattern FORMCYCLE_PLACEHOLDER = Pattern.compile("^\\[%\\s*\\$.+%]$");
	private static final boolean AS_FIELD_NAME = false;
	private static final Set<String> EXPECTED_TRUST_LEVELS = Set.of("STORK-QAA-Level-1", "STORK-QAA-Level-2", "STORK-QAA-Level-3",
			"STORK-QAA-Level-4");
	static final String TEXT_KEY = "text";
	static final String KEY_POSTKORB_FORM = "PostboxId";
	static final String KEY_TRUST_LEVEL = "TrustLevel";

	private static final Predicate<FormFieldMetaData> IS_SELECT_FIELD = fieldMetaData -> fieldMetaData.getType() == EFormFieldTyp.SELECT;
	private static final Predicate<Map.Entry<String, FormFieldMetaData>> IS_UPLOAD_FIELD = e -> e.getValue().getType() == EFormFieldTyp.UPLOAD;

	private final IWorkflowExecutionEnvironmentData environmentData;
	private final String organisationsEinheitId;

	public FormData readFormData() {

		var formFields = buildFormData();
		Set<String> uploadFieldNames = getUploadFieldNames();
		var formDataBuilder = FormData.builder()
				.formId(environmentData.getProject().getId())
				.formName(getFormName())
				.organisationsEinheitId(organisationsEinheitId)
				.fieldData(formFields)
				.attachmentNames(getAttachmentNames(uploadFieldNames))
				.attachmentUuids(getAttachmentUuids(uploadFieldNames));
		addServiceKonto(formDataBuilder);
		return formDataBuilder.build();
	}

	List<FormNode> buildFormData() {
		var formNodes = getFormStructure();
		return setSubNodeValues(formNodes);
	}

	List<FormNode> setSubNodeValues(List<FormNode> formNodes) {
		List<FormNode> nodesWithValues = new ArrayList<>();
		for (FormNode formNode : formNodes) {
			if (isServiceKontoRelatedNode(formNode)) {
				continue;
			}
			formNode.setValues(getValues(formNode.getName()));
			setSubNodeValues(formNode.getNestedElements());
			nodesWithValues.add(formNode);
		}
		return nodesWithValues;
	}

	boolean isServiceKontoRelatedNode(FormNode formNode) {
		return KEY_POSTKORB_FORM.equals(formNode.getName()) || KEY_TRUST_LEVEL.equals(formNode.getName());
	}

	List<String> getValues(String fieldName) {
		List<String[]> fieldValues = getFormDataAdapter().getFieldValues(fieldName);
		if (isNull(fieldValues)) {
			return Collections.emptyList();
		}
		return fieldValues.stream()
				.flatMap(Arrays::stream)
				.map(value -> getValueAsText(fieldName, value))
				.collect(Collectors.toList());
	}

	String getValueAsText(String fieldName, String value) {
		if (isNull(value) || value.isBlank()) {
			return value;
		}
		return Optional.ofNullable(getFormDataAdapter().getFieldMetaData(fieldName, AS_FIELD_NAME))
				.filter(IS_SELECT_FIELD)
				.map(fieldMetaData -> fieldMetaData.getOptionsMap().get(value))
				.map(valueWithText -> valueWithText.get(TEXT_KEY))
				.orElse(value);
	}

	Set<String> getAttachmentNames(Set<String> uploadFieldNames) {
		return uploadFieldNames.stream()
				.map(getFormDataAdapter()::getFieldValues)
				.filter(Objects::nonNull)
				.flatMap(List::stream)
				.flatMap(Arrays::stream)
				.collect(Collectors.toSet());
	}

	Set<String> getAttachmentUuids(Set<String> uploadFieldNames) {
		var pattern = String.format("(%s)\\.uuid", String.join("|", uploadFieldNames));
		return getFormDataAdapter().getFieldValuesMap().entrySet().stream()
				.filter(e -> e.getKey().matches(pattern))
				.flatMap(e -> e.getValue().stream())
				.flatMap(Arrays::stream)
				.collect(Collectors.toSet());
	}

	Set<String> getUploadFieldNames() {
		return getFormDataAdapter().getFieldMetaDataMap().entrySet().stream()
				.filter(IS_UPLOAD_FIELD).map(Map.Entry::getKey).collect(Collectors.toSet());
	}

	void addServiceKonto(FormData.FormDataBuilder formDataBuilder) {
		try {
			getPostkorbId().map(this::buildServiceKonto).ifPresent(formDataBuilder::serviceKonto);
		} catch (UnexpectedTrustLevelException e) {
			formDataBuilder.warning(e.getWarning());
		}
	}

	Optional<String> getPostkorbId() {
		return Optional.ofNullable(getFormDataAdapter().getFieldValues(KEY_POSTKORB_FORM))
				.flatMap(this::extractPostkorbId);
	}

	Optional<String> extractPostkorbId(List<String[]> values) {
		return values.stream().flatMap(Stream::of).filter(this::isValidPostkorbId).findFirst();
	}

	boolean isValidPostkorbId(String postkorbId) {
		return isNotBlank(postkorbId) && !FORMCYCLE_PLACEHOLDER.matcher(postkorbId).matches();
	}

	ServiceKonto buildServiceKonto(String postkorbId) {
		return ServiceKonto.builder().postkorbId(postkorbId).trustLevel(getTrustLevel()).build();
	}

	String getTrustLevel() {
		var trustLevel = Optional.ofNullable(getFormDataAdapter().getFieldValues(KEY_TRUST_LEVEL)).flatMap(this::extractTrustLevel)
				.orElseThrow(() -> new UnexpectedTrustLevelException(
						"TrustLevel not found. Please ensure that the 'Vertrauensniveau' element (name='TrustLevel') is added to the Bürger Services template. BayernID user account is not connected"));
		if (!EXPECTED_TRUST_LEVELS.contains(trustLevel)) {
			throw new UnexpectedTrustLevelException(
					String.format("TrustLevel has an unexpected value '%s'. BayernID user account is not connected", trustLevel));
		}
		return trustLevel;
	}

	Optional<String> extractTrustLevel(List<String[]> values) {
		return values.stream().flatMap(Stream::of).filter(StringUtils::isNotBlank).findFirst();
	}

	String getFormName() {
		return environmentData.getProject().getTitel();
	}

	IFormDataAdapter getFormDataAdapter() {
		return environmentData.getFormDataAdapter();
	}

	List<FormNode> getFormStructure() {
		return createFormStructureParser().parse();
	}

	FormStructureParser createFormStructureParser() {
		return new FormStructureParser(environmentData.getXForm().getXItems());
	}
}
