/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle.validation;

import java.lang.reflect.InvocationTargetException;
import java.util.Optional;
import java.util.Set;

import javax.validation.ConstraintValidator;
import javax.validation.ConstraintValidatorContext;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class ConditionalNotEmptyValidator implements ConstraintValidator<ConditionalNotEmpty, Object> {

	private static final Logger LOG = LoggerFactory.getLogger(ConditionalNotEmptyValidator.class);

	private String conditionalProperty;
	private Set<String> requiredProperties;
	private Set<String> conditionalValues;
	private String constraintViolationMessage;

	@Override
	public void initialize(ConditionalNotEmpty constraint) {
		conditionalProperty = constraint.dependingOn();
		conditionalValues = Set.of(constraint.withValues());
		requiredProperties = Set.of(constraint.requiredProperties());
		constraintViolationMessage = constraint.message();
	}

	@Override
	public boolean isValid(Object object, ConstraintValidatorContext context) {
		Optional<String> conditionalPropertyValue = getPropertyValue(object, conditionalProperty);
		if (conditionalPropertyValue.isEmpty()) {
			return false;
		}
		return conditionalPropertyValue.filter(conditionalValues::contains).map(value -> validateRequiredProperties(object, context)).orElse(true);
	}

	boolean validateRequiredProperties(Object object, ConstraintValidatorContext context) {
		for (String property : requiredProperties) {
			var propertyValue = getPropertyValue(object, property);
			if (propertyValue.isEmpty()) {
				addConstraintViolation(context, property);
				return false;
			}
		}
		return true;

	}
	Optional<String> getPropertyValue(Object bean, String propertyName) {
		try {
			return Optional.ofNullable(BeanUtils.getProperty(bean, propertyName)).filter(StringUtils::isNotBlank);
		} catch (IllegalAccessException | InvocationTargetException | NoSuchMethodException e) {
			LOG.error("Error getting property value for property '{}' from bean '{}'", propertyName, bean, e);
		}
		return Optional.empty();
	}

	void addConstraintViolation(ConstraintValidatorContext context, String propertyName) {
		context.disableDefaultConstraintViolation();
		context.buildConstraintViolationWithTemplate(constraintViolationMessage).addPropertyNode(propertyName).addConstraintViolation();
	}
}
