/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle;

import static de.ozgcloud.formcycle.OzgCloudFormDataHttpClient.*;
import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.lang.reflect.Field;
import java.util.List;
import java.util.Optional;

import org.apache.http.HttpHost;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.protocol.HttpClientContext;
import org.apache.http.conn.routing.HttpRoutePlanner;
import org.apache.http.impl.client.BasicCredentialsProvider;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.protocol.HttpContext;
import org.apache.logging.log4j.core.util.ReflectionUtil;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.NullSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.util.ReflectionUtils;

import de.ozgcloud.eingang.formcycle.FormCycleFormData;
import de.ozgcloud.formcycle.attachment.FormcycleAttachment;
import de.ozgcloud.formcycle.attachment.FormcycleAttachmentTestFactory;
import de.ozgcloud.formcycle.client.OzgCloudResponseHandler;
import de.ozgcloud.formcycle.errorhandling.OzgCloudRequestException;
import lombok.SneakyThrows;

class OzgCloudFormDataHttpClientTest {

	@Spy
	@InjectMocks
	private OzgCloudFormDataHttpClient client;

	@Mock
	private SystemPropertiesProvider systemPropertiesProvider;
	@Mock
	private FormCycleFormData formData;
	@Mock
	private CloseableHttpClient httpClient;
	@Mock
	private OzgCloudConfig ozgCloudConfig;
	@Mock
	private PortParser portParser;
	@Mock
	private OzgCloudResponseHandler responseHandler;

	private List<FormcycleAttachment> attachments;
	private List<FormcycleAttachment> representations;

	@Nested
	class TestSendData {

		@Mock
		private HttpPost postRequest;
		@Mock
		private HttpContext httpContext;

		@SneakyThrows
		@BeforeEach
		void setup() {
			attachments = List.of(FormcycleAttachmentTestFactory.create());
			representations = List.of(FormcycleAttachmentTestFactory.create());
			doReturn(httpClient).when(client).createCloseableClient();
			doReturn(postRequest).when(client).buildPostRequest(any(), any(), any());
			doReturn(httpContext).when(client).getContext();
		}

		@SneakyThrows
		@Test
		void shouldCallBuildRequest() {
			client.send(formData, attachments, representations);

			verify(client).buildPostRequest(formData, attachments, representations);
		}

		@SneakyThrows
		@Test
		void shouldSendData() {
			when(httpClient.execute(any(), any(OzgCloudResponseHandler.class), any())).thenReturn(FormCycleConfirmationResponseTestFactory.create());

			var response = client.send(formData, attachments, representations);

			assertThat(response).isEqualTo(FormCycleConfirmationResponseTestFactory.create());
		}

		@Nested
		class TestThrowException {

			@Mock
			private IOException ioException;

			private OzgCloudRequestException exception = spy(new OzgCloudRequestException("message"));

			@SneakyThrows
			@Test
			void shouldWrapIOException() {
				when(httpClient.execute(any(), any(OzgCloudResponseHandler.class), any(HttpContext.class))).thenThrow(ioException);

				var thrownException = assertThrows(OzgCloudRequestException.class, () -> client.send(formData, attachments, representations));
				assertThat(thrownException.getCause()).isEqualTo(ioException);
			}
		}
	}

	@Nested
	class TestClientConfig {

		@Test
		@DisplayName("should create http with connect timeout")
		void shouldCreateConfigConnectTimeout() {
			var httpClientConfig = client.createHttpClientConfig();

			assertThat(httpClientConfig.getConnectTimeout()).isEqualTo(HTTPCLIENT_TIMEOUT);
		}

		@Test
		@DisplayName("should create http with connection request timeout")
		void shouldCreateConfigRequestTimeout() {
			var httpClientConfig = client.createHttpClientConfig();

			assertThat(httpClientConfig.getConnectionRequestTimeout()).isEqualTo(HTTPCLIENT_TIMEOUT);
		}

		@Test
		@DisplayName("should create http with socket timeout")
		void shouldCreateConfigSocketTimeout() {
			var httpClientConfig = client.createHttpClientConfig();

			assertThat(httpClientConfig.getSocketTimeout()).isEqualTo(HTTPCLIENT_TIMEOUT);
		}

	}

	@Nested
	class TestCreateCloseableClient {

		@Test
		void shouldCallCreateHttpClientConfig() {
			client.createCloseableClient();

			verify(client).createHttpClientConfig();
		}

		@Test
		void shouldCallAddProxy() {
			client.createCloseableClient();

			verify(client).addProxy(any());
		}

		@Test
		void shouldCreateClient() {
			var httpClient = client.createCloseableClient();

			assertThat(httpClient).isNotNull();
		}
	}

	@Nested
	class TestAddProxy {

		@Mock
		private HttpClientBuilder httpClientBuilder;
		@Mock
		private ProxyConfig proxyConfig;

		@Test
		void shouldCallConfigureProxy() {
			doReturn(proxyConfig).when(client).getProxyConfig();
			doNothing().when(client).configureProxy(any());

			client.addProxy(httpClientBuilder);

			verify(client).configureProxy(httpClientBuilder);
		}
	}

	@Nested
	class TestConfigureProxy {

		@Mock
		private HttpClientBuilder httpClientBuilder;
		@Mock
		private HttpHost proxyHost;
		@Mock
		private BasicCredentialsProvider credentialsProvider;
		@Mock
		private HttpRoutePlanner routePlanner;

		@BeforeEach
		void setup() {
			doReturn(routePlanner).when(client).createHttpRouterPlanner();
			doReturn(credentialsProvider).when(client).getCredentialsProvider();
		}

		@Test
		void shouldCallSetRouterPlanner() {

			client.configureProxy(httpClientBuilder);

			verify(httpClientBuilder).setRoutePlanner(routePlanner);
		}

		@Test
		void shouldCallSetCredentialProvider() {

			client.configureProxy(httpClientBuilder);

			verify(httpClientBuilder).setDefaultCredentialsProvider(credentialsProvider);
		}
	}

	@Nested
	class TestGetProxyConfig {

		@Test
		void shouldReturnFromOzgCloudConfig() {
			when(client.getProxyConfig()).thenReturn(OzgCloudConfigTestFactory.proxyConfig);

			var proxyConfig = client.getProxyConfig();

			assertThat(proxyConfig).isEqualTo(OzgCloudConfigTestFactory.proxyConfig);
		}

		@Test
		void shouldCallLoadSystemProxyConfig() {
			client.getProxyConfig();

			verify(client).loadSystemProxyConfig();
		}

		@Test
		void shouldReturnNullWhenNoProxy() {
			//			client = createOzgClient(ProxyConfigTestFactory.createEmpty());
			doReturn(Optional.empty()).when(client).loadSystemProxyConfig();

			var proxyConfig = client.getProxyConfig();

			assertThat(proxyConfig).isNull();
		}

	}

	@Nested
	class TestLoadSystemProxyConfig {

		@Test
		void shouldCallLoadHttpsProxyConfig() {
			doReturn(ProxyConfigTestFactory.create()).when(client).loadHttpsProxyConfig();

			client.loadSystemProxyConfig();

			verify(client).loadHttpsProxyConfig();
		}

		@Test
		@DisplayName("should not load http settings when https is set")
		void shouldNotLoadHttpWhenHttps() {
			doReturn(ProxyConfigTestFactory.create()).when(client).loadHttpsProxyConfig();

			client.loadSystemProxyConfig();

			verify(client, never()).loadHttpProxyConfig();
		}

		@Test
		void shouldCallLoadHttpProxyConfig() {
			doReturn(null).when(client).loadHttpsProxyConfig();

			client.loadSystemProxyConfig();

			verify(client).loadHttpProxyConfig();
		}
	}

	@Nested
	class TestLoadHttpsProxyConfig {

		@BeforeEach
		void setup() {
			when(systemPropertiesProvider.getHttpsProxyHost()).thenReturn(PropertiesTestFactory.PROXY_HOST);
		}

		@ParameterizedTest(name = "should return null when \"{0}\"")
		@NullSource
		@ValueSource(strings = { "", "  " })
		void shouldReturnNull(String property) {
			lenient().when(systemPropertiesProvider.getHttpsProxyHost()).thenReturn(property);

			var proxyConfig = client.loadHttpsProxyConfig();

			assertThat(proxyConfig).isNull();
		}

		@Test
		void shouldSetProxyHost() {
			when(systemPropertiesProvider.getHttpsProxyHost()).thenReturn(PropertiesTestFactory.PROXY_HOST);

			var proxyConfig = client.loadHttpsProxyConfig();

			assertThat(proxyConfig.getHost()).isEqualTo(PropertiesTestFactory.PROXY_HOST);
		}

		@Test
		void shouldSetProxyPort() {
			when(portParser.parse(any())).thenReturn(ProxyConfigTestFactory.PROXY_PORT);

			var proxyConfig = client.loadHttpsProxyConfig();

			assertThat(proxyConfig.getPort()).isEqualTo(ProxyConfigTestFactory.PROXY_PORT);
		}

		@Test
		void shouldSetProxyUser() {
			when(systemPropertiesProvider.getHttpsProxyUser()).thenReturn(PropertiesTestFactory.PROXY_USER);

			var proxyConfig = client.loadHttpsProxyConfig();

			assertThat(proxyConfig.getUser()).isEqualTo(PropertiesTestFactory.PROXY_USER);
		}

		@Test
		void shouldSetProxyPassword() {
			when(systemPropertiesProvider.getHttpsProxyPassword()).thenReturn(PropertiesTestFactory.PROXY_PASSWORD);

			var proxyConfig = client.loadHttpsProxyConfig();

			assertThat(proxyConfig.getPassword()).isEqualTo(PropertiesTestFactory.PROXY_PASSWORD);
		}
	}

	@Nested
	class TestLoadHttpProxyConfig {

		@BeforeEach
		void setup() {
			when(systemPropertiesProvider.getHttpProxyHost()).thenReturn(PropertiesTestFactory.PROXY_HOST);
		}

		@ParameterizedTest(name = "should return null when \"{0}\"")
		@NullSource
		@ValueSource(strings = { "", "  " })
		void shouldReturnNull(String property) {
			lenient().when(systemPropertiesProvider.getHttpProxyHost()).thenReturn(property);

			var proxyConfig = client.loadHttpProxyConfig();

			assertThat(proxyConfig).isNull();
		}

		@Test
		void shouldSetProxyHost() {

			var proxyConfig = client.loadHttpProxyConfig();

			assertThat(proxyConfig.getHost()).isEqualTo(PropertiesTestFactory.PROXY_HOST);
		}

		@Test
		void shouldSetProxyPort() {
			when(portParser.parse(any())).thenReturn(ProxyConfigTestFactory.PROXY_PORT);

			var proxyConfig = client.loadHttpProxyConfig();

			assertThat(proxyConfig.getPort()).isEqualTo(ProxyConfigTestFactory.PROXY_PORT);
		}

		@Test
		void shouldSetProxyUser() {
			when(systemPropertiesProvider.getHttpProxyUser()).thenReturn(PropertiesTestFactory.PROXY_USER);

			var proxyConfig = client.loadHttpProxyConfig();

			assertThat(proxyConfig.getUser()).isEqualTo(PropertiesTestFactory.PROXY_USER);
		}

		@Test
		void shouldSetProxyPassword() {
			when(systemPropertiesProvider.getHttpProxyPassword()).thenReturn(PropertiesTestFactory.PROXY_PASSWORD);

			var proxyConfig = client.loadHttpProxyConfig();

			assertThat(proxyConfig.getPassword()).isEqualTo(PropertiesTestFactory.PROXY_PASSWORD);
		}
	}

	@Nested
	class TestGetContext {

		@Mock
		private HttpClientContext context;

		@Test
		void shouldReturnCreatedContext() {
			doReturn(context).when(client).createContext();

			var result = client.getContext();

			assertThat(result).isEqualTo(context);
		}

		@Test
		void shouldReturnCachedContext() {
			ReflectionUtil.setFieldValue(getField("context"), client, context);

			var result = client.getContext();

			assertThat(result).isEqualTo(context);
		}

		@Nested
		class TestCreateContext {

			@Mock
			private BasicCredentialsProvider credentialsProvider;
			@Mock
			private ProxyConfig proxyConfig;

			@Test
			void shouldSetCredentialsProvider() {
				doReturn(proxyConfig).when(client).getProxyConfig();
				doReturn(credentialsProvider).when(client).getCredentialsProvider();

				var result = client.createContext();

				assertThat(result.getCredentialsProvider()).isEqualTo(credentialsProvider);
			}

			@Test
			void shouldReturnEmptyContext() {
				doReturn(null).when(client).getProxyConfig();

				client.createContext();

				verify(client, never()).getCredentialsProvider();
			}
		}
	}

	private Field getField(String name) {
		return ReflectionUtils.findField(OzgCloudFormDataHttpClient.class, name);
	}
}