/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle;

import static de.ozgcloud.formcycle.PropertiesTestFactory.*;
import static java.util.Objects.*;
import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Properties;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.CsvSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.formcycle.errorhandling.TechnicalException;
import de.xima.fc.interfaces.plugin.lifecycle.IPluginInitializeData;

class OzgPluginPropertiesSupplierTest {
	@Spy
	@InjectMocks
	private OzgPluginPropertiesSupplier mapper;

	@Mock
	private IPluginInitializeData pluginData;
	@Mock
	private PortParser portParser;

	private Properties properties = PropertiesTestFactory.create();

	@Nested
	class TestMapProxyProperty {

		@Mock
		private ProxyConfig proxyConfig;

		@Test
		void shouldCallValidateProxyConfig() {
			doReturn(properties).when(pluginData).getProperties();

			mapper.getProxyProperties();

			verify(mapper).validateProxyConfig(properties);
		}

		@Test
		void shouldReturnProxyConfig() {
			doReturn(properties).when(pluginData).getProperties();
			doNothing().when(mapper).validateProxyConfig(any());
			doReturn(proxyConfig).when(mapper).buildProxyConfig(any());

			var result = mapper.getProxyProperties();

			assertThat(result).isPresent().get().isEqualTo(proxyConfig);
		}

		@Test
		void shouldReturnEmptyWhenNoProxy() {
			doReturn(PropertiesTestFactory.createEmpty()).when(pluginData).getProperties();

			var result = mapper.getProxyProperties();

			assertThat(result).isEmpty();
		}
	}

	@Nested
	class TestBuildProxyConfig {

		@Test
		void shouldSetHost() {
			var result = mapper.buildProxyConfig(properties);

			assertThat(result.getHost()).isEqualTo(PROXY_HOST);
		}

		@Test
		void shouldSetPort() {
			doReturn(1).when(mapper).getPort(any());

			var result = mapper.buildProxyConfig(properties);

			assertThat(result.getPort()).isEqualTo(1);
		}

		@Test
		void shouldSetUser() {
			var result = mapper.buildProxyConfig(properties);

			assertThat(result.getUser()).isEqualTo(PROXY_USER);
		}

		@Test
		void shouldSetPassword() {
			var result = mapper.buildProxyConfig(properties);

			assertThat(result.getPassword()).isEqualTo(PROXY_PASSWORD);
		}

	}

	@Nested
	class TestValidateProxyConfig {

		@DisplayName(value = "should fail when")
		@ParameterizedTest(name = "host: \"{0}\" and port: \"{1}\"")
		@CsvSource({ "host, ", ", 1", "'',1", "host,' '" })
		void shouldFailWhen(String host, String port) {
			var properties = createProperties(host, port);

			assertThrows(TechnicalException.class, () -> mapper.validateProxyConfig(properties));
		}

		@Test
		void shouldValidate() {
			assertDoesNotThrow(() -> mapper.validateProxyConfig(properties));
		}

		private Properties createProperties(String host, String port) {
			var properties = new Properties();
			if (nonNull(host)) {
				properties.setProperty(ProxyConfig.KEY_PROXY_HOST, host);
			}
			if (nonNull(port)) {
				properties.setProperty(ProxyConfig.KEY_PROXY_PORT, port);
			}
			return properties;
		}
	}

	@Nested
	class TestGetPort {

		@Test
		void shouldReturnNull() {
			var result = mapper.getPort(PropertiesTestFactory.createEmpty());

			assertThat(result).isNull();
		}

		@Test
		void shouldCallParsePort() {
			mapper.getPort(properties);

			verify(portParser).parse(PROXY_PORT);
		}
	}
}