/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle.client;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.InputStream;
import java.nio.charset.StandardCharsets;

import org.apache.commons.io.IOUtils;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.StatusLine;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import com.fasterxml.jackson.databind.ObjectMapper;

import de.ozgcloud.eingang.formcycle.FormCycleConfirmationResponse;
import de.ozgcloud.formcycle.errorhandling.OzgCloudRequestException;
import lombok.SneakyThrows;

class OzgCloudResponseHandlerTest {

	@Spy
	@InjectMocks
	private OzgCloudResponseHandler responseHandler;

	@Mock
	private ObjectMapper objectMapper;

	@Mock
	private HttpResponse httpResponse;

	@Nested
	class TestHandleResponse {

		@Mock
		private HttpEntity httpEntity;

		@SneakyThrows
		@BeforeEach
		void setUp() {
			doNothing().when(responseHandler).checkResponseStatus(any());
		}

		@SneakyThrows
		@Test
		void shouldCheckResponseStatus() {
			when(httpResponse.getEntity()).thenReturn(httpEntity);
			when(httpEntity.getContent()).thenReturn(InputStream.nullInputStream());

			responseHandler.handleResponse(httpResponse);

			verify(responseHandler).checkResponseStatus(httpResponse);
		}

		@SneakyThrows
		@Test
		void shouldReturnVorgangNummer() {
			var expectedConformationResponse = FormCycleConfirmationResponse.newBuilder().setVorgangNummer("123").build();
			doReturn(expectedConformationResponse).when(responseHandler).buildResponse(httpResponse);

			var response = responseHandler.handleResponse(httpResponse);

			assertThat(response).isEqualTo(expectedConformationResponse);
		}

	}

	@Nested
	class TestCheckResponseStatus {

		@Mock
		private StatusLine statusLine;

		private final OzgCloudErrorDto errorDto = new OzgCloudErrorDto();
		private final String message = "message";
		private final String exceptionId = "exception-id";

		@BeforeEach
		void setUp() {
			errorDto.setMessage(message);
			errorDto.setExceptionId(exceptionId);
		}

		@Test
		void shouldPassWhenOk() {
			when(statusLine.getStatusCode()).thenReturn(HttpStatus.SC_OK);
			when(httpResponse.getStatusLine()).thenReturn(statusLine);

			assertDoesNotThrow(() -> responseHandler.checkResponseStatus(httpResponse));
		}

		@Test
		void shouldThrowExceptionWhenError() {
			when(statusLine.getStatusCode()).thenReturn(HttpStatus.SC_BAD_REQUEST);
			when(httpResponse.getStatusLine()).thenReturn(statusLine);
			doReturn(errorDto).when(responseHandler).getOzgCloudError(any());

			var ozgCloudRequestException = assertThrows(OzgCloudRequestException.class, () -> responseHandler.checkResponseStatus(httpResponse));

			assertThat(ozgCloudRequestException.getMessage()).startsWith(message);
			assertThat(ozgCloudRequestException.getExceptionId()).isEqualTo(exceptionId);
		}

		@Test
		void shouldThrowExceptionWhenNotOk() {
			when(statusLine.getStatusCode()).thenReturn(HttpStatus.SC_NO_CONTENT);
			when(httpResponse.getStatusLine()).thenReturn(statusLine);
			doReturn(message).when(responseHandler).buildFullMessage(any());

			var ozgCloudRequestException = assertThrows(OzgCloudRequestException.class, () -> responseHandler.checkResponseStatus(httpResponse));

			assertThat(ozgCloudRequestException.getMessage()).startsWith(message);
		}
	}

	@Nested
	class TestBuildFullMessage {

		@Mock
		private HttpEntity httpEntity;

		@Mock
		private StatusLine statusLine;

		private String messageBody = "message-body";

		@SneakyThrows
		@BeforeEach
		void setUp() {
			when(statusLine.toString()).thenReturn(String.valueOf(HttpStatus.SC_BAD_REQUEST));
			when(httpResponse.getStatusLine()).thenReturn(statusLine);
			var inputStream = IOUtils.toInputStream(messageBody, StandardCharsets.UTF_8);
			when(httpEntity.getContent()).thenReturn(inputStream);
			when(httpResponse.getEntity()).thenReturn(httpEntity);
		}

		@Test
		void shouldBuildMessage() {
			var expectedMessage = String.format(OzgCloudResponseHandler.MESSAGE, HttpStatus.SC_BAD_REQUEST) + "\n" + messageBody;

			var message = responseHandler.buildFullMessage(httpResponse);

			assertThat(message).isEqualTo(expectedMessage);
		}

	}
}