/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle.errorhandling;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.formcycle.OzgCloudConfig;
import de.ozgcloud.formcycle.attachment.FormcycleAttachment;
import de.xima.fc.interfaces.workflow.params.INodeThrewExceptionBuilder;

class NodeThrewExceptionFactoryTest {

	private static final String EXCEPTION_ID = "exception-id";

	@Spy
	@InjectMocks
	private NodeThrewExceptionFactory exceptionFactory;

	@Mock
	private INodeThrewExceptionBuilder exceptionBuilder;
	@Mock
	private OzgCloudConfig pluginConfig;

	@Captor
	private ArgumentCaptor<String> stringCaptor;
	@Captor
	private ArgumentCaptor<Map<String, String>> errorDataCaptor;
	@Captor
	private ArgumentCaptor<Exception> errorCauseCaptor;
	@Mock
	private FormcycleAttachment expectedAttachments;

	@DisplayName("Create internal exception")
	@Nested
	class TestCreateInternalException {

		@BeforeEach
		void init() {
			when(exceptionBuilder.error(any(), any())).thenReturn(exceptionBuilder);
			when(exceptionBuilder.cause(any())).thenReturn(exceptionBuilder);
		}

		@DisplayName("form TechnicalException")
		@Nested
		class TestFromTechnicalException {

			@Mock
			private TechnicalException exception;

			@BeforeEach
			void setup() {
				when(exceptionBuilder.message(any())).thenReturn(exceptionBuilder);
				when(exception.getExceptionId()).thenReturn(EXCEPTION_ID);
			}

			@Test
			void shouldAddErrorCode() {
				when(exceptionBuilder.error(stringCaptor.capture(), any())).thenReturn(exceptionBuilder);

				exceptionFactory.createInternalException(exception);

				assertThat(stringCaptor.getValue()).isEqualTo(OzgPluginError.INTERNAL_ERROR.name());
			}

			@Test
			void shouldAddExceptionId() {
				when(exceptionBuilder.error(any(), errorDataCaptor.capture())).thenReturn(exceptionBuilder);

				exceptionFactory.createInternalException(exception);

				var errorData = errorDataCaptor.getValue();
				assertThat(errorData).containsEntry(NodeThrewExceptionFactory.KEY_EXCEPTION_ID, EXCEPTION_ID);
			}

			@Test
			void shouldCallFormatWithPluginConfig() {
				var errorMessage = "error message";
				when(exception.getMessage()).thenReturn(errorMessage);

				exceptionFactory.createInternalException(exception);

				verify(exceptionFactory).formatWithPluginConfig(errorMessage);
			}

			@Test
			void shouldAddExceptionMessage() {
				var errorMessage = "error message";
				doReturn(errorMessage).when(exceptionFactory).formatWithPluginConfig(any());

				exceptionFactory.createInternalException(exception);

				verify(exceptionBuilder).message(stringCaptor.capture());
				assertThat(stringCaptor.getValue()).isEqualTo(errorMessage);
			}

			@Test
			void shouldAddCause() {
				when(exceptionBuilder.cause(errorCauseCaptor.capture())).thenReturn(exceptionBuilder);

				exceptionFactory.createInternalException(exception);

				assertThat(errorCauseCaptor.getValue()).isEqualTo(exception);
			}
		}

		@DisplayName("from message and exception")
		@Nested
		class TestFromException {

			private static final String EXCEPTION_MESSAGE = "exception message";
			@Mock
			private Exception exception;

			@BeforeEach
			void init() {
				when(exceptionBuilder.message(any())).thenReturn(exceptionBuilder);
			}

			@Test
			void shouldAddErrorCode() {
				createException();

				verify(exceptionBuilder).error(stringCaptor.capture(), any());
				assertThat(stringCaptor.getValue()).isEqualTo(OzgPluginError.INTERNAL_ERROR.name());
			}

			@Test
			void shouldAddExceptionId() {
				doReturn(Map.of(NodeThrewExceptionFactory.KEY_EXCEPTION_ID, EXCEPTION_ID)).when(exceptionFactory).createDataMap(any());

				createException();

				verify(exceptionBuilder).error(any(), errorDataCaptor.capture());
				var errorData = errorDataCaptor.getValue();
				assertThat(errorData).containsEntry(NodeThrewExceptionFactory.KEY_EXCEPTION_ID, EXCEPTION_ID);
			}

			@Test
			void shouldCallFormatWithExceptionId() {
				try(var exceptionUtilMock = mockStatic(ExceptionUtil.class)) {
					exceptionUtilMock.when(ExceptionUtil::createExceptionId).thenReturn(EXCEPTION_ID);

					createException();

					exceptionUtilMock.verify(() -> ExceptionUtil.formatMessageWithExceptionId(EXCEPTION_MESSAGE));
				}
			}

			@Test
			void shouldCallFormatWithPluginConfig() {
				try(var exceptionUtilMock = mockStatic(ExceptionUtil.class)) {
					exceptionUtilMock.when(ExceptionUtil::createExceptionId).thenReturn(EXCEPTION_ID);
					exceptionUtilMock.when(() -> ExceptionUtil.formatMessageWithExceptionId(any())).thenReturn(EXCEPTION_MESSAGE);

					createException();

					verify(exceptionFactory).formatWithPluginConfig(EXCEPTION_MESSAGE);
				}
			}

			@Test
			void shouldAddExceptionMessage() {
				doReturn(EXCEPTION_MESSAGE).when(exceptionFactory).formatWithPluginConfig(any());

				createException();

				verify(exceptionBuilder).message(stringCaptor.capture());
				assertThat(stringCaptor.getValue()).isEqualTo(EXCEPTION_MESSAGE);
			}

			@Test
			void shouldAddCause() {
				when(exceptionBuilder.cause(errorCauseCaptor.capture())).thenReturn(exceptionBuilder);

				createException();

				assertThat(errorCauseCaptor.getValue()).isEqualTo(exception);
			}

			private void createException() {
				exceptionFactory.createInternalException(EXCEPTION_MESSAGE, exception);
			}
		}
	}

	@Nested
	class TestFromOzgCloudRequestException {
		@Mock
		private OzgCloudRequestException exception;

		@BeforeEach
		void setup() {
			when(exceptionBuilder.error(any(), any())).thenReturn(exceptionBuilder);
			when(exceptionBuilder.cause(any())).thenReturn(exceptionBuilder);
			when(exceptionBuilder.message(any())).thenReturn(exceptionBuilder);
			when(exception.getExceptionId()).thenReturn(EXCEPTION_ID);
		}

		@Test
		void shouldAddErrorCode() {
			when(exceptionBuilder.error(stringCaptor.capture(), any())).thenReturn(exceptionBuilder);

			createException();

			assertThat(stringCaptor.getValue()).isEqualTo(OzgPluginError.OZGCLOUD_REQUEST_FAILED.name());
		}

		@Test
		void shouldAddExceptionId() {
			when(exceptionBuilder.error(any(), errorDataCaptor.capture())).thenReturn(exceptionBuilder);

			createException();

			var errorData = errorDataCaptor.getValue();
			assertThat(errorData).containsEntry(NodeThrewExceptionFactory.KEY_EXCEPTION_ID, EXCEPTION_ID);
		}

		@Test
		void shouldCallFormatWithPluginConfig() {
			var errorMessage = "error message";
			when(exception.getMessage()).thenReturn(errorMessage);

			createException();

			verify(exceptionFactory).formatWithPluginConfig(errorMessage);
		}

		@Test
		void shouldSetFormattedMessage() {
			var errorMessage = "error message";
			doReturn(errorMessage).when(exceptionFactory).formatWithPluginConfig(any());

			createException();

			verify(exceptionBuilder).message(stringCaptor.capture());
			assertThat(stringCaptor.getValue()).isEqualTo(errorMessage);
		}

		@Test
		void shouldAddCause() {
			createException();

			verify(exceptionBuilder).cause(errorCauseCaptor.capture());
			assertThat(errorCauseCaptor.getValue()).isEqualTo(exception);
		}

		private void createException() {
			exceptionFactory.createOzgCloudRequestException(exception);
		}
	}

	@Nested
	class TestCreateDataMap {

		@Test
		void shouldAddExceptionId() {
			var dataMap = exceptionFactory.createDataMap(EXCEPTION_ID);

			assertThat(dataMap).containsEntry(NodeThrewExceptionFactory.KEY_EXCEPTION_ID, EXCEPTION_ID);
		}

		@Test
		void shouldHandleNullValue() {
			var dataMap = exceptionFactory.createDataMap(null);

			assertThat(dataMap.get(NodeThrewExceptionFactory.KEY_EXCEPTION_ID)).isNotEmpty();
		}

	}
}