/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle.formdata;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;

import org.assertj.core.api.InstanceOfAssertFactories;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.NullAndEmptySource;
import org.junit.jupiter.params.provider.NullSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.formcycle.errorhandling.UnexpectedTrustLevelException;
import de.ozgcloud.formcycle.errorhandling.Warning;
import de.xima.fc.entities.Projekt;
import de.xima.fc.form.common.items.XItem;
import de.xima.fc.form.common.models.IXForm;
import de.xima.fc.interfaces.form.IFormDataAdapter;
import de.xima.fc.interfaces.workflow.execution.IWorkflowExecutionEnvironmentData;
import de.xima.fc.mdl.enums.EFormFieldTyp;
import de.xima.fc.mdl.fdv.form.FormFieldMetaData;

class PluginFormDataAdapterTest {

	private static final String[] VALUES = new String[] { "0", "1" };

	@Spy
	@InjectMocks
	private PluginFormDataAdapter adapter;

	@Mock
	private IWorkflowExecutionEnvironmentData environmentData;
	@Mock
	private IFormDataAdapter formDataAdapter;

	@Nested
	class TestReadFormData {

		private static final String UPLOAD_FIELD_NAME = "upload";
		private final String ATTACHMENT_UUID = UUID.randomUUID().toString();

		@Mock
		private Projekt project;

		private FormNode formNode = FormNodeTestFactory.create();

		@BeforeEach
		void setup() {
			adapter = spy(new PluginFormDataAdapter(environmentData, FormDataTestFactory.ORGANISATIONS_EINHEIT_ID));
			doReturn(List.of(formNode)).when(adapter).buildFormData();
			doReturn(Set.of(UPLOAD_FIELD_NAME)).when(adapter).getUploadFieldNames();
			doReturn(FormDataTestFactory.FORM_NAME).when(adapter).getFormName();
			doReturn(Set.of(FormDataTestFactory.ATTACHMENT_ID)).when(adapter).getAttachmentNames(any());
			doReturn(Set.of(ATTACHMENT_UUID)).when(adapter).getAttachmentUuids(any());
			doNothing().when(adapter).addServiceKonto(any());
			when(project.getId()).thenReturn(FormDataTestFactory.FORM_ID);
			when(environmentData.getProject()).thenReturn(project);
		}

		@Test
		void shouldCallBuildFormData() {
			adapter.readFormData();

			verify(adapter).buildFormData();
		}

		@Test
		void shouldCallGetUploadFieldNames() {
			adapter.readFormData();

			verify(adapter).getUploadFieldNames();
		}

		@Test
		void shouldSetFormId() {
			var result = adapter.readFormData();

			assertThat(result.getFormId()).isEqualTo(FormDataTestFactory.FORM_ID);
		}

		@Test
		void shouldSetFormName() {
			var result = adapter.readFormData();

			assertThat(result.getFormName()).isEqualTo(FormDataTestFactory.FORM_NAME);
		}

		@Test
		void shouldSetOrganisationseinheitenId() {
			var result = adapter.readFormData();

			assertThat(result.getOrganisationsEinheitId()).isEqualTo(FormDataTestFactory.ORGANISATIONS_EINHEIT_ID);
		}

		@Test
		void shouldSetFieldData() {
			var result = adapter.readFormData();

			assertThat(result.getFieldData()).containsExactly(formNode);
		}

		@Test
		void shouldSetAttachmentName() {
			var result = adapter.readFormData();

			assertThat(result.getAttachmentNames()).containsExactly(FormDataTestFactory.ATTACHMENT_ID);
		}

		@Test
		void shouldSetAttachmentUuids() {
			var result = adapter.readFormData();

			assertThat(result.getAttachmentUuids()).containsExactly(ATTACHMENT_UUID);
		}

		@Test
		void shouldCallAddServiceKonto() {
			adapter.readFormData();

			verify(adapter).addServiceKonto(any());
		}

	}

	@Nested
	class TestBuildFormData {

		private FormNode formNode = FormNodeTestFactory.create();

		@Captor
		private ArgumentCaptor<List<FormNode>> formNodeCaptor;

		@BeforeEach
		void setup() {
		}

		@Test
		@DisplayName("should call getFormStructure")
		void shouldCallGetFormStructure() {
			doReturn(List.of(FormNodeTestFactory.create())).when(adapter).getFormStructure();
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);

			adapter.buildFormData();

			verify(adapter).getFormStructure();
		}

		@Test
		void shouldCallSetSubNodeValues() {
			doReturn(List.of(formNode)).when(adapter).getFormStructure();
			// don't call recursion
			doReturn(true).when(adapter).isServiceKontoRelatedNode(any());

			adapter.buildFormData();

			verify(adapter).setSubNodeValues(formNodeCaptor.capture());
			assertThat(formNodeCaptor.getValue()).containsExactly(formNode);
		}

		@Test
		void shouldReturnFormNodes() {
			doReturn(List.of(FormNodeTestFactory.create())).when(adapter).getFormStructure();
			var expectedNodes = List.of(FormNodeTestFactory.create());
			doReturn(expectedNodes).when(adapter).setSubNodeValues(any());

			var formNodes = adapter.buildFormData();

			assertThat(formNodes).isEqualTo(expectedNodes);
		}
	}

	@Nested
	class TestSetSubNodeValues {

		@Test
		void shouldSetValue() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);
			when(formDataAdapter.getFieldValues(anyString())).thenReturn(PluginDataTestFactory.createFieldValues(FormDataTestFactory.FIELD_VALUE_1));
			doReturn(FormDataTestFactory.FIELD_VALUE_1).when(adapter).getValueAsText(anyString(), anyString());
			FormNode formNode = FormNodeTestFactory.create();

			var result = adapter.setSubNodeValues(List.of(formNode));

			assertThat(result).hasSize(1).first().extracting(FormNode::getValues, InstanceOfAssertFactories.LIST)
					.containsExactly(FormDataTestFactory.FIELD_VALUE_1);
		}

		@Test
		void shouldCallSetSubNodeValues() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);
			FormNode formNode = FormNodeTestFactory.create();

			adapter.setSubNodeValues(List.of(formNode));

			verify(adapter, times(2)).setSubNodeValues(any());
		}

		@Test
		void shouldCallIsServiceKontoRelatedNode() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);
			FormNode formNode = FormNodeTestFactory.create();

			adapter.setSubNodeValues(List.of(formNode));

			verify(adapter).isServiceKontoRelatedNode(formNode);
		}

		@Test
		void shouldReturnEmptyWhenServiceKontoRelatedNode() {
			doReturn(true).when(adapter).isServiceKontoRelatedNode(any());

			var result = adapter.setSubNodeValues(List.of(FormNodeTestFactory.create()));

			assertThat(result).isEmpty();
		}

		@Nested
		class TestIsServiceKontoRelatedNode {

			@Test
			void shouldSkipPostkorbIdNode() {
				var formNode = FormNodeTestFactory.createBuilder().name(PluginFormDataAdapter.KEY_POSTKORB_FORM).build();

				var formFieldsOut = adapter.setSubNodeValues(List.of(formNode));

				assertThat(formFieldsOut).isEmpty();
			}

			@Test
			void shouldSkipTrustLevelNode() {
				var formNode = FormNodeTestFactory.createBuilder().name(PluginFormDataAdapter.KEY_TRUST_LEVEL).build();

				var formFieldsOut = adapter.setSubNodeValues(List.of(formNode));

				assertThat(formFieldsOut).isEmpty();
			}
		}

	}

	@Nested
	class TestIsServiceKontoRelatedNode {

		@Test
		void shouldReturnTrueWhenPostkorbId() {
			var formNode = FormNodeTestFactory.createBuilder().name(PluginFormDataAdapter.KEY_POSTKORB_FORM).build();

			var result = adapter.isServiceKontoRelatedNode(formNode);

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnTrueWhenTrustLevel() {
			var formNode = FormNodeTestFactory.createBuilder().name(PluginFormDataAdapter.KEY_TRUST_LEVEL).build();

			var result = adapter.isServiceKontoRelatedNode(formNode);

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalseWhenNotServiceKontoRelated() {
			var result = adapter.isServiceKontoRelatedNode(FormNodeTestFactory.create());

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestGetValues {

		@BeforeEach
		void init() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);
		}

		@Test
		void shouldCallGetFormDataAdapter() {
			adapter.getValues(PluginDataTestFactory.FIELD_KEY);

			verify(environmentData).getFormDataAdapter();
		}

		@Test
		void shouldCallGetFieldValues() {
			adapter.getValues(PluginDataTestFactory.FIELD_KEY);

			verify(formDataAdapter).getFieldValues(PluginDataTestFactory.FIELD_KEY);
		}

		@Test
		void shouldReturnEmptyWhenNoData() {
			when(formDataAdapter.getFieldValues(anyString())).thenReturn(Collections.emptyList());

			var result = adapter.getValues(PluginDataTestFactory.FIELD_KEY);

			assertThat(result).isEmpty();
		}

		@Test
		void shouldCallGetValueAsText() {
			var values2 = new String[] { "2" };
			when(formDataAdapter.getFieldValues(anyString())).thenReturn(List.of(VALUES, values2));

			adapter.getValues(PluginDataTestFactory.FIELD_KEY);

			verify(adapter).getValueAsText(PluginDataTestFactory.FIELD_KEY, "0");
			verify(adapter).getValueAsText(PluginDataTestFactory.FIELD_KEY, "1");
			verify(adapter).getValueAsText(PluginDataTestFactory.FIELD_KEY, "2");
		}

		@Test
		void shouldReturnValues() {
			when(formDataAdapter.getFieldValues(anyString())).thenReturn(createValuesList(VALUES));

			var result = adapter.getValues(PluginDataTestFactory.FIELD_KEY);

			assertThat(result).containsExactly(VALUES);
		}
	}

	@Nested
	class TestGetValueAsText {

		@Mock
		private IFormDataAdapter formDataAdapter;
		@Mock
		private FormFieldMetaData formFieldMetaData;

		@DisplayName("should return input when")
		@ParameterizedTest(name = "value = \"{0}\"")
		@NullAndEmptySource
		void shouldReturnWhenValueNull(String sourceValue) {
			var result = adapter.getValueAsText(PluginDataTestFactory.FIELD_KEY, sourceValue);

			assertThat(result).isEqualTo(sourceValue);
		}

		@Test
		void shouldCallGetFormDataAdapter() {
			doReturn(formDataAdapter).when(adapter).getFormDataAdapter();

			adapter.getValueAsText(PluginDataTestFactory.FIELD_KEY, PluginDataTestFactory.FIELD_VALUE);

			verify(adapter).getFormDataAdapter();
		}

		@Test
		void shouldCallGetFieldMetaData() {
			doReturn(formDataAdapter).when(adapter).getFormDataAdapter();

			adapter.getValueAsText(PluginDataTestFactory.FIELD_KEY, PluginDataTestFactory.FIELD_VALUE);

			verify(formDataAdapter).getFieldMetaData(PluginDataTestFactory.FIELD_KEY, false);
		}

		@DisplayName("should filter out when")
		@ParameterizedTest(name = "field type is {0}")
		@EnumSource(value = EFormFieldTyp.class, names = { "SELECT" }, mode = EnumSource.Mode.EXCLUDE)
		void shouldFilterOut(EFormFieldTyp fieldTyp) {
			doReturn(formDataAdapter).when(adapter).getFormDataAdapter();
			when(formFieldMetaData.getType()).thenReturn(fieldTyp);
			when(formDataAdapter.getFieldMetaData(anyString(), anyBoolean())).thenReturn(formFieldMetaData);
			var sourceValue = PluginDataTestFactory.FIELD_VALUE;

			var result = adapter.getValueAsText(PluginDataTestFactory.FIELD_KEY, sourceValue);

			assertThat(result).isEqualTo(sourceValue);
		}

		@Test
		void shouldReturnSelectFieldValues() {
			doReturn(formDataAdapter).when(adapter).getFormDataAdapter();
			when(formFieldMetaData.getType()).thenReturn(EFormFieldTyp.SELECT);
			when(formFieldMetaData.getOptionsMap()).thenReturn(PluginDataTestFactory.createOptionsMap());
			when(formDataAdapter.getFieldMetaData(any(), anyBoolean())).thenReturn(formFieldMetaData);

			var result = adapter.getValueAsText(PluginDataTestFactory.FIELD_KEY, PluginDataTestFactory.FIELD_VALUE);

			assertThat(result).isEqualTo(PluginDataTestFactory.TEXT_MAP_VALUE);
		}

		@Test
		void shouldReturnInputValueWhenNoTextValue() {
			doReturn(formDataAdapter).when(adapter).getFormDataAdapter();
			when(formFieldMetaData.getType()).thenReturn(EFormFieldTyp.SELECT);
			when(formFieldMetaData.getOptionsMap()).thenReturn(
					PluginDataTestFactory.createOptionsMapWithTextMapKey(PluginDataTestFactory.TITLE_MAP_KEY));
			when(formDataAdapter.getFieldMetaData(any(), anyBoolean())).thenReturn(formFieldMetaData);

			var result = adapter.getValueAsText(PluginDataTestFactory.FIELD_KEY, PluginDataTestFactory.FIELD_VALUE);

			assertThat(result).isEqualTo(PluginDataTestFactory.FIELD_VALUE);
		}

		@Test
		@DisplayName("should return same value when there's no metadata")
		void shouldReturnWhenNoMetadata() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);

			var result = adapter.getValueAsText(PluginDataTestFactory.FIELD_KEY, PluginDataTestFactory.FIELD_VALUE);

			assertThat(result).isEqualTo(PluginDataTestFactory.FIELD_VALUE);

		}
	}

	@Nested
	@DisplayName("Get Attachment Names")
	class TestAddAttachmentNames {

		private static final String FIELD_VALUE_1 = "value1";
		private static final String FIELD_VALUE_2 = "value2";

		@BeforeEach
		void init() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);
		}

		@Test
		@DisplayName("should skip upload field when no content was uploaded")
		void shouldSkipWhenNoData() {
			when(formDataAdapter.getFieldValues(anyString())).thenReturn(null);

			var attachmentNames = adapter.getAttachmentNames(Set.of(PluginDataTestFactory.FIELD_KEY));

			assertThat(attachmentNames).isEmpty();
		}

		@Test
		@DisplayName("should add all attachments when multiple elements have the same name")
		void shouldGetAllAttachments1() {
			when(formDataAdapter.getFieldValues(anyString())).thenReturn(
					PluginDataTestFactory.createFieldValuesWhenEqualUploadName(FIELD_VALUE_1, FIELD_VALUE_2));

			var attachmentNames = adapter.getAttachmentNames(Set.of(PluginDataTestFactory.FIELD_KEY));

			assertThat(attachmentNames).containsExactlyInAnyOrder(FIELD_VALUE_1, FIELD_VALUE_2);
		}

		@Test
		@DisplayName("should add all attachments")
		void shouldGetAllAttachments2() {
			when(formDataAdapter.getFieldValues(anyString())).thenReturn(PluginDataTestFactory.createFieldValues(FIELD_VALUE_1, FIELD_VALUE_2));

			var attachmentNames = adapter.getAttachmentNames(Set.of(PluginDataTestFactory.FIELD_KEY));

			assertThat(attachmentNames).containsExactlyInAnyOrder(FIELD_VALUE_1, FIELD_VALUE_2);
		}
	}

	@Nested
	@DisplayName("Get Attachment UUIDs")
	class TestAttachmentUuids {

		private static final String FIELD_VALUE_1 = "value1";

		private Map<String, List<String[]>> fieldValuesMap = new HashMap<>();

		@BeforeEach
		void setup() {
			fieldValuesMap.put(PluginDataTestFactory.FIELD_KEY, PluginDataTestFactory.createFieldValues(FIELD_VALUE_1));
			fieldValuesMap.put(PluginDataTestFactory.UUID_FIELD_KEY, PluginDataTestFactory.createFieldValues(PluginDataTestFactory.UUID_VALUE));
			when(formDataAdapter.getFieldValuesMap()).thenReturn(fieldValuesMap);
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);
		}

		@Test
		void shouldGetAttachmentUuids() {
			var attachmentUuids = adapter.getAttachmentUuids(Set.of(PluginDataTestFactory.FIELD_KEY));

			assertThat(attachmentUuids).hasSize(1).first().isEqualTo(PluginDataTestFactory.UUID_VALUE);
		}

		@Test
		@DisplayName("should add all attachments when upload elements have different names")
		void shouldGetAllAttachmentUuidsFromList() {
			var uuid2 = UUID.randomUUID().toString();
			fieldValuesMap.put("field2.uuid", PluginDataTestFactory.createFieldValues(PluginDataTestFactory.UUID_VALUE, uuid2));

			var attachmentUuids = adapter.getAttachmentUuids(Set.of(PluginDataTestFactory.FIELD_KEY, "field2"));

			assertThat(attachmentUuids).hasSize(2).containsExactlyInAnyOrder(PluginDataTestFactory.UUID_VALUE, uuid2);
		}

		@Test
		@DisplayName("should add all attachments when upload elements have same name")
		void shouldGetAllAttachmentUuidsFromArray() {
			var uuid2 = UUID.randomUUID().toString();
			fieldValuesMap.put("field2.uuid", PluginDataTestFactory.createFieldValuesWhenEqualUploadName(PluginDataTestFactory.UUID_VALUE, uuid2));

			var attachmentUuids = adapter.getAttachmentUuids(Set.of(PluginDataTestFactory.FIELD_KEY, "field2"));

			assertThat(attachmentUuids).hasSize(2).containsExactlyInAnyOrder(PluginDataTestFactory.UUID_VALUE, uuid2);
		}
	}

	@Nested
	class TestGetUploadFieldNames {

		@Mock
		private FormFieldMetaData formFieldMetaData;

		@BeforeEach
		void init() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);
		}

		@Test
		@DisplayName("should skip not upload fields")
		void shouldSkipFields() {
			when(formFieldMetaData.getType()).thenReturn(EFormFieldTyp.TEXTFIELD);
			when(formDataAdapter.getFieldMetaDataMap()).thenReturn(Map.of(PluginDataTestFactory.FIELD_KEY, formFieldMetaData));

			var uploadFieldNames = adapter.getUploadFieldNames();

			assertThat(uploadFieldNames).isEmpty();
		}

		@Test
		void shouldReturnWhenNoData() {
			when(formDataAdapter.getFieldMetaDataMap()).thenReturn(Collections.emptyMap());

			var uploadFieldNames = adapter.getUploadFieldNames();

			assertThat(uploadFieldNames).isEmpty();
		}

		@Test
		@DisplayName("should return names of upload fields")
		void shouldReturnFieldNames() {
			when(formFieldMetaData.getType()).thenReturn(EFormFieldTyp.UPLOAD);
			when(formDataAdapter.getFieldMetaDataMap()).thenReturn(Map.of(PluginDataTestFactory.FIELD_KEY, formFieldMetaData));

			var uploadFieldNames = adapter.getUploadFieldNames();

			assertThat(uploadFieldNames).containsExactlyInAnyOrder(PluginDataTestFactory.FIELD_KEY);
		}
	}

	@Nested
	class TestAddServiceKonto {

		private FormData.FormDataBuilder formDataBuilder = FormData.builder();

		@Test
		void shouldCallGetPostkorbId() {
			doReturn(Optional.empty()).when(adapter).getPostkorbId();

			adapter.addServiceKonto(formDataBuilder);

			verify(adapter).getPostkorbId();
		}

		@Test
		void shouldCallBuildServiceKonto() {
			doReturn(Optional.of(ServiceKontoTestFactory.POSTKORB_ID)).when(adapter).getPostkorbId();
			doReturn(ServiceKontoTestFactory.create()).when(adapter).buildServiceKonto(any());

			adapter.addServiceKonto(formDataBuilder);

			verify(adapter).buildServiceKonto(ServiceKontoTestFactory.POSTKORB_ID);
		}

		@Test
		void shouldSetServiceKonto() {
			doReturn(Optional.of(ServiceKontoTestFactory.POSTKORB_ID)).when(adapter).getPostkorbId();
			var expectedServiceKonto = ServiceKontoTestFactory.create();
			doReturn(expectedServiceKonto).when(adapter).buildServiceKonto(any());

			adapter.addServiceKonto(formDataBuilder);

			assertThat(formDataBuilder.build().getServiceKonto()).contains(expectedServiceKonto);
		}

		@Test
		void shouldAddWarningByException() {
			var exceptionMessage = "message";
			doThrow(new UnexpectedTrustLevelException(exceptionMessage)).when(adapter).getPostkorbId();

			adapter.addServiceKonto(formDataBuilder);

			assertThat(formDataBuilder.build().getWarnings()).hasSize(1).first().extracting(Warning::getMessage, InstanceOfAssertFactories.STRING)
					.contains(exceptionMessage);
		}
	}

	@Nested
	class TestGetPostkorbId {

		@BeforeEach
		void setup() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);
		}

		@Test
		void shouldMapPostkorbId() {
			var values = new ArrayList<String[]>();
			values.add(new String[] { ServiceKontoTestFactory.POSTKORB_ID });
			when(formDataAdapter.getFieldValues(anyString())).thenReturn(values);

			Optional<String> postkorbId = adapter.getPostkorbId();

			assertThat(postkorbId).contains(ServiceKontoTestFactory.POSTKORB_ID);
		}

		@Test
		void shouldNotMapOnEmptyPostkorbId() {
			when(formDataAdapter.getFieldValues(anyString())).thenReturn(Collections.emptyList());

			Optional<String> postkorbId = adapter.getPostkorbId();

			assertThat(postkorbId).isEmpty();
		}
	}

	@Nested
	class TestPostkorbIdValidation {

		@DisplayName("should return FALSE when")
		@ParameterizedTest(name = "postkorbId is \"{0}\"")
		@NullSource
		@ValueSource(strings = { "", "[%$id%]", "[% $id%]", "[%$id %]", "[% $id %]" })
		void shouldFilterOutEmptyValues(String postkorbId) {
			var isValid = adapter.isValidPostkorbId(postkorbId);

			assertThat(isValid).isFalse();
		}
	}

	@Nested
	class TestBuildServiceKonto {

		private static final String TRUST_LEVEL = "trust-level";

		@BeforeEach
		void init() {
			doReturn(TRUST_LEVEL).when(adapter).getTrustLevel();
		}

		@Test
		void shouldSetPostkorbId() {
			var result = buildServiceKonto();

			assertThat(result.getPostkorbId()).isEqualTo(ServiceKontoTestFactory.POSTKORB_ID);
		}

		@Test
		void shouldCallGetTrustLevel() {
			buildServiceKonto();

			verify(adapter).getTrustLevel();
		}

		@Test
		void shouldSetTrustLevel() {
			var result = buildServiceKonto();

			assertThat(result.getTrustLevel()).isEqualTo(TRUST_LEVEL);
		}

		private ServiceKonto buildServiceKonto() {
			return adapter.buildServiceKonto(ServiceKontoTestFactory.POSTKORB_ID);
		}
	}

	@Nested
	class TestGetTrustLevel {

		@BeforeEach
		void setup() {
			when(environmentData.getFormDataAdapter()).thenReturn(formDataAdapter);
		}

		@Test
		void shouldThrowExceptionWhenNoSuchField() {
			assertThrows(UnexpectedTrustLevelException.class, () -> adapter.getTrustLevel());
		}

		@Test
		void shouldCallExtractTrustLevel() {
			when(formDataAdapter.getFieldValues(anyString())).thenReturn(createValuesList());

			try {
				adapter.getTrustLevel();
			} catch (UnexpectedTrustLevelException e) {
			}

			verify(adapter).extractTrustLevel(any());
		}

		@Test
		void shouldThrowExceptionWhenNoTrustLevel() {
			doReturn(Optional.empty()).when(adapter).extractTrustLevel(any());

			assertThrows(UnexpectedTrustLevelException.class, () -> adapter.getTrustLevel());
		}

		@Test
		void shouldThrowExceptionWhenNotExpectedTrustLevel() {
			doReturn(Optional.of("unexpected")).when(adapter).extractTrustLevel(any());

			assertThrows(UnexpectedTrustLevelException.class, () -> adapter.getTrustLevel());
		}

		@Test
		void shouldReturnTrustLevel() {
			doReturn(Optional.of(ServiceKontoTestFactory.TRUST_LEVEL)).when(adapter).extractTrustLevel(any());

			var trustLevel = adapter.getTrustLevel();

			assertThat(trustLevel).isEqualTo(ServiceKontoTestFactory.TRUST_LEVEL);
		}
	}

	@Nested
	class TestExtractTrustLevel {

		@Test
		void shouldReturnEmptyWhenNoValues() {
			var result = adapter.extractTrustLevel(Collections.emptyList());

			assertThat(result).isEmpty();
		}

		@DisplayName("should return empty when")
		@ParameterizedTest(name = "field values are {0}")
		@NullAndEmptySource
		void shouldReturnEmptyWhen(String value) {
			var result = adapter.extractTrustLevel(createValuesList(value));

			assertThat(result).isEmpty();
		}

		@Test
		void shouldReturnValue() {
			var values = new ArrayList<String[]>();
			values.add(new String[] { ServiceKontoTestFactory.TRUST_LEVEL });

			var result = adapter.extractTrustLevel(values);

			assertThat(result).contains(ServiceKontoTestFactory.TRUST_LEVEL);
		}
	}

	@Nested
	class TestGetFormName {

		@Mock
		private Projekt projekt;

		@BeforeEach
		void setup() {
			when(environmentData.getProject()).thenReturn(projekt);
			when(projekt.getTitel()).thenReturn(FormDataTestFactory.FORM_NAME);
		}

		@Test
		void shouldCallGetProject() {
			adapter.getFormName();

			verify(environmentData).getProject();
		}

		@Test
		void shouldCallGetTitel() {
			adapter.getFormName();

			verify(projekt).getTitel();
		}

		@Test
		void shouldReturnValue() {
			var formName = adapter.getFormName();

			assertThat(formName).isEqualTo(FormDataTestFactory.FORM_NAME);
		}
	}

	@Nested
	class TestCreateFormParser {

		@Mock
		private FormStructureParser formStructureParser;

		@BeforeEach
		void setup() {
			doReturn(formStructureParser).when(adapter).createFormStructureParser();
		}

		@Test
		void shouldCallCreateFormParser() {
			adapter.getFormStructure();

			verify(adapter).createFormStructureParser();
		}

		@Test
		void shouldCallParse() {
			adapter.getFormStructure();

			verify(formStructureParser).parse();
		}
	}

	@Nested
	class TestGetFormStructure {

		@Mock
		private IXForm xForm;
		@Mock
		private XItem xItem;

		private Map<String, XItem> xItems;

		@BeforeEach
		void init() {
			when(environmentData.getXForm()).thenReturn(xForm);
			when(xForm.getXItems()).thenReturn(xItems);
			xItems = Map.of("key", xItem);
		}

		@Test
		void shouldCallGetXForm() {
			adapter.createFormStructureParser();

			verify(environmentData).getXForm();
		}

		@Test
		void shouldCallGetXItems() {
			adapter.createFormStructureParser();

			verify(xForm).getXItems();
		}

	}

	List<String[]> createValuesList(String... value) {
		var values = new ArrayList<String[]>();
		values.add(value);
		return values;
	}

}