/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle.validation;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.lang.reflect.InvocationTargetException;
import java.util.Optional;
import java.util.Set;

import javax.validation.ConstraintValidatorContext;
import javax.validation.ConstraintValidatorContext.ConstraintViolationBuilder;
import javax.validation.ConstraintValidatorContext.ConstraintViolationBuilder.NodeBuilderCustomizableContext;

import org.apache.commons.beanutils.BeanUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.NullAndEmptySource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Mockito;
import org.mockito.Spy;

import lombok.SneakyThrows;

class ConditionalNotEmptyValidatorTest {

	private static final String CONDITIONAL_VALUE = "conditionalValue";
	private static final String REQUIRED_PROPERTY = "requiredProperty";
	private static final String CONDITIONAL_PROPERTY = "conditionalProperty";

	@Spy
	@InjectMocks
	private ConditionalNotEmptyValidator validator;

	@Mock
	private ConditionalNotEmpty constraint;
	@Mock
	private Object bean;

	@Nested
	class TestInitialize {

		@BeforeEach
		void setup() {
			when(constraint.withValues()).thenReturn(new String[] { CONDITIONAL_VALUE });
			when(constraint.requiredProperties()).thenReturn(new String[] { REQUIRED_PROPERTY });
		}

		@Test
		void shouldInitializeConditionalProperty() {
			when(constraint.dependingOn()).thenReturn(CONDITIONAL_PROPERTY);

			validator.initialize(constraint);

			assertThat(validator).extracting("conditionalProperty").isEqualTo(CONDITIONAL_PROPERTY);
		}

		@Test
		void shouldInitializeConditionalValues() {
			validator.initialize(constraint);

			assertThat(validator).extracting("conditionalValues").isEqualTo(Set.of(CONDITIONAL_VALUE));
		}

		@Test
		void shouldInitializeRequiredProperties() {
			validator.initialize(constraint);

			assertThat(validator).extracting("requiredProperties").isEqualTo(Set.of(REQUIRED_PROPERTY));
		}

		@Test
		void shouldInitializeConstraintViolationMessage() {
			var message = "message";
			when(constraint.message()).thenReturn(message);

			validator.initialize(constraint);

			assertThat(validator).extracting("constraintViolationMessage").isEqualTo(message);
		}
	}

	@Nested
	class TestIsValid {

		@Mock
		private ConstraintValidatorContext context;

		@SneakyThrows
		@BeforeEach
		private void setup() {
			when(constraint.dependingOn()).thenReturn(CONDITIONAL_PROPERTY);
			when(constraint.withValues()).thenReturn(new String[] { CONDITIONAL_VALUE });
			when(constraint.requiredProperties()).thenReturn(new String[] { REQUIRED_PROPERTY });
			validator.initialize(constraint);
		}

		@Test
		void shouldRejectIfNoPropertyValue() {
			doReturn(Optional.empty()).when(validator).getPropertyValue(any(), any());

			var isValid = validator.isValid(bean, context);

			assertFalse(isValid);
		}

		@Test
		void shouldRejectIfRequiredPropertyValueIsEmpty() {
			doReturn(Optional.empty()).when(validator).getPropertyValue(any(), any());
			doReturn(Optional.of(CONDITIONAL_VALUE)).when(validator).getPropertyValue(any(), eq(CONDITIONAL_PROPERTY));
			doNothing().when(validator).addConstraintViolation(any(), any());

			var isValid = validator.isValid(bean, context);

			assertFalse(isValid);
			verify(validator).validateRequiredProperties(bean, context);
		}

		@Test
		void shouldApproveByDifferentConditionalValue() {
			doReturn(Optional.of("another_value")).when(validator).getPropertyValue(any(), eq(CONDITIONAL_PROPERTY));

			var isValid = validator.isValid(bean, context);

			assertTrue(isValid);
		}
	}

	@Nested
	class TestValidateRequiredProperties {

		@Mock
		private ConstraintValidatorContext context;
		@Mock
		private ConstraintViolationBuilder constraintViolationBuilder;
		@Mock
		private NodeBuilderCustomizableContext nodeBuilderCustomizableContext;

		@SneakyThrows
		@BeforeEach
		private void setup() {
			when(constraint.dependingOn()).thenReturn(CONDITIONAL_PROPERTY);
			when(constraint.withValues()).thenReturn(new String[] { CONDITIONAL_VALUE });
			when(constraint.requiredProperties()).thenReturn(new String[] { REQUIRED_PROPERTY });
			validator.initialize(constraint);
		}

		@Test
		void shouldApproveIfNoRequiredProperty() {
			when(constraint.requiredProperties()).thenReturn(new String[] {});
			validator.initialize(constraint);

			var isValid = validator.validateRequiredProperties(bean, context);

			assertTrue(isValid);
		}

		@Test
		void shouldApproveRequiredProperty() {
			doReturn(Optional.of("value")).when(validator).getPropertyValue(any(), any());

			var isValid = validator.validateRequiredProperties(bean, context);

			assertTrue(isValid);
		}

		@Test
		void shouldRejectRequiredProperty() {
			doReturn(Optional.empty()).when(validator).getPropertyValue(any(), any());
			when(context.buildConstraintViolationWithTemplate(any())).thenReturn(constraintViolationBuilder);
			when(constraintViolationBuilder.addPropertyNode(any())).thenReturn(nodeBuilderCustomizableContext);

			var isValid = validator.validateRequiredProperties(bean, context);

			assertFalse(isValid);
		}

		@Test
		void shouldCallAddConstraintViolation() {
			doReturn(Optional.empty()).when(validator).getPropertyValue(any(), any());
			when(context.buildConstraintViolationWithTemplate(any())).thenReturn(constraintViolationBuilder);
			when(constraintViolationBuilder.addPropertyNode(any())).thenReturn(nodeBuilderCustomizableContext);

			validator.validateRequiredProperties(bean, context);

			verify(validator).addConstraintViolation(context, REQUIRED_PROPERTY);
		}
	}

	@Nested
	class TestGetPropertyValue {

		@Test
		void shouldReturnEmptyIfIllegalAccessException() {
			try (MockedStatic<BeanUtils> utilities = Mockito.mockStatic(BeanUtils.class)) {
				utilities.when(() -> BeanUtils.getProperty(any(), any())).thenThrow(new IllegalAccessException());

				var propertyValue = validator.getPropertyValue(bean, REQUIRED_PROPERTY);

				assertThat(propertyValue).isEmpty();
			}
		}

		@Test
		void shouldReturnEmptyIfInvocationTargetException() {
			try (MockedStatic<BeanUtils> utilities = Mockito.mockStatic(BeanUtils.class)) {
				utilities.when(() -> BeanUtils.getProperty(any(), any())).thenThrow(new InvocationTargetException(null));

				var propertyValue = validator.getPropertyValue(bean, REQUIRED_PROPERTY);

				assertTrue(propertyValue.isEmpty());
			}
		}

		@Test
		void shouldReturnEmptyIfNoSuchMethodException() {
			try (MockedStatic<BeanUtils> utilities = Mockito.mockStatic(BeanUtils.class)) {
				utilities.when(() -> BeanUtils.getProperty(any(), any())).thenThrow(new NoSuchMethodException());

				var propertyValue = validator.getPropertyValue(bean, REQUIRED_PROPERTY);

				assertTrue(propertyValue.isEmpty());
			}
		}

		@ParameterizedTest(name = "should return empty if \"{0}\"")
		@NullAndEmptySource
		@ValueSource(strings = { " " })
		void shouldReturnEmptyIfPropertyValueIsNull(String propertyValue) {
			try (MockedStatic<BeanUtils> utilities = Mockito.mockStatic(BeanUtils.class)) {
				utilities.when(() -> BeanUtils.getProperty(any(), any())).thenReturn(propertyValue);

				var resultValue = validator.getPropertyValue(bean, REQUIRED_PROPERTY);

				assertThat(resultValue).isEmpty();
			}
		}

		@Test
		void shouldReturnPropertyValue() {
			try (MockedStatic<BeanUtils> utilities = Mockito.mockStatic(BeanUtils.class)) {
				utilities.when(() -> BeanUtils.getProperty(any(), any())).thenReturn("value");

				var resultValue = validator.getPropertyValue(bean, REQUIRED_PROPERTY);

				assertThat(resultValue).contains("value");
			}
		}
	}

	@Nested
	class TestAddConstraintViolation {

		@Mock
		private ConstraintValidatorContext context;
		@Mock
		private ConstraintViolationBuilder constraintViolationBuilder;
		@Mock
		private NodeBuilderCustomizableContext nodeBuilderCustomizableContext;

		@BeforeEach
		void setup() {
			when(context.buildConstraintViolationWithTemplate(any())).thenReturn(constraintViolationBuilder);
			when(constraintViolationBuilder.addPropertyNode(any())).thenReturn(nodeBuilderCustomizableContext);
		}

		@Test
		void shouldCallDisableDefaultConstraintViolation() {
			validator.addConstraintViolation(context, REQUIRED_PROPERTY);

			verify(context).disableDefaultConstraintViolation();
		}

		@Test
		void shouldCallBuildConstraintViolationWithTemplate() {
			var message = "message";
			initValidator(message);

			validator.addConstraintViolation(context, REQUIRED_PROPERTY);

			verify(context).buildConstraintViolationWithTemplate(message);
		}

		@Test
		void shuldCallAddPropertyNode() {
			validator.addConstraintViolation(context, REQUIRED_PROPERTY);

			verify(constraintViolationBuilder).addPropertyNode(REQUIRED_PROPERTY);
		}

		@Test
		void shouldCallAddConstraintViolation() {
			validator.addConstraintViolation(context, REQUIRED_PROPERTY);

			verify(nodeBuilderCustomizableContext).addConstraintViolation();
		}

		private void initValidator(String message) {
			when(constraint.message()).thenReturn(message);
			when(constraint.withValues()).thenReturn(new String[] { CONDITIONAL_VALUE });
			when(constraint.requiredProperties()).thenReturn(new String[] { REQUIRED_PROPERTY });
			validator.initialize(constraint);
		}
	}
}